#include <string>
#include <iostream>
#include <fstream>
#include <cstring>
#include "bmp.h"
using namespace std;

// read a single short integer from a file in PC format (low byte comes first)
bool read_PC_short(ifstream &f, short &num)
{

    char byte;
    short byteshifted(0);
    num = 0;
    // Read the low bytes first, shift 8 more bits each time
    for (int i = 0; i <= 1; i++) {
        if (!f.get(byte)) {
            cout << "read_bitmap_file: error reading short from file" << endl;
            return false;
        }
        byteshifted = byte << (i*8);
        num|=byteshifted;
    }
    return true;
}

// Read a single integer from a file in PC format (lowest to highest bytes)
bool read_PC_int(ifstream &f, int &num)
{
    char byte;
    int byteshifted(0);
    num = 0;
    for (int i = 0; i <= 3; i++) {
        if (!f.get(byte)) {
            cout << "read_bitmap_file: error reading int from file" << endl;
            return false;
        }
        byteshifted = byte << (i*8);
        num|=byteshifted;
    }
    return true;
}

// Each pixel gets written 3 times - one each for RGB.  This simulates gray
bool read_three_copies(ifstream &bmpfile, unsigned char &pixel)
{
    int k;
    unsigned char red, green, blue;
    // read 3 copies of each pixel, one for each color
    if (!bmpfile.get(red) || !bmpfile.get(green) || !bmpfile.get(blue)) {
            return false;
    }
    if (red != blue || blue != green) {
        return false;
    }
    pixel = red;
    return true;
}

// reads a grayscale .bmp file.  Pass it a file name, it reads the file
// and passes back output values width and height, and returns an
// unsigned char * pointing to the image data.
// This array contains the image data, and was dynamically allocated by
// this function.  The caller is responsible for deleting (releasing)
// this memory.  The file is assumed to be a grayscale image (value for
// Red, Green and Blue must be the same).
// If the read fails at any point, the function returns 0 (NULL)
unsigned char * read_bitmap_file(string bmpfn, int &w, int &h)
{
    ifstream bmpfile(bmpfn.c_str(), ios::in |ios::binary); // input file
    bmfh_struct  bmpfh;   // file header
    bmih_struct bmih;     // info header
    unsigned char *data;

    // check that file opened ok
    if (!bmpfile.is_open()) {
        cout << "Read bitmap file: error opening file " << bmpfn << endl;
        return false;
    }

    // Read the file header
    bmpfile.get(bmpfh.bfType[0]);
    bmpfile.get(bmpfh.bfType[1]);
    if (!read_PC_int(bmpfile,bmpfh.bfSize)) return false;
    if (!read_PC_short(bmpfile,bmpfh.bfReserved1)) return false;
    if (!read_PC_short(bmpfile,bmpfh.bfReserved2)) return false;
    if (!read_PC_int(bmpfile,bmpfh.bfOffBits)) return false;

    // check all values in the file header
    if ((bmpfh.bfType[0]!= BMF_DEFAULT_TYPE[0]) ||
    	(bmpfh.bfType[1]!= BMF_DEFAULT_TYPE[1])) {
        cout << "Read bitmap file: Bad file type" << endl;
        return false;
    }
    if (bmpfh.bfReserved1 != 0 || bmpfh.bfReserved2 != 0 ||
        bmpfh.bfOffBits != BMF_DEFAULT_SIZE+BMIH_DEFAULT_SIZE) {
        cout << "Read bitmap file: Bad file format" << endl;
        return false;
    }

    // read information header from file
    if (!read_PC_int(bmpfile, bmih.biSize)) return false;
    if (!read_PC_int(bmpfile, bmih.biWidth)) return false;
    if (!read_PC_int(bmpfile, bmih.biHeight)) return false;
    if (!read_PC_short(bmpfile, bmih.biPlanes)) return false;
    if (!read_PC_short(bmpfile, bmih.biBitCount)) return false;
    if (!read_PC_int(bmpfile, bmih.biCompression)) return false;
    if (!read_PC_int(bmpfile, bmih.biSizeImage)) return false;
    if (!read_PC_int(bmpfile, bmih.biXPelsPerMeter)) return false;
    if (!read_PC_int(bmpfile, bmih.biYPelsPerMeter)) return false;
    if (!read_PC_int(bmpfile, bmih.biClrUsed)) return false;
    if (!read_PC_int(bmpfile, bmih.biClrImportant)) return false;
    
    // check values in information header
    if (bmih.biSize != BMIH_DEFAULT_SIZE) {
        cout << "Read bitmap file: Bad information header size" << endl;
        return false;
    }
    w = bmih.biWidth;
    h = bmih.biHeight;   // Height of image in pixels
    if (bmpfh.bfSize != (BMF_DEFAULT_SIZE+BMIH_DEFAULT_SIZE+w*h*3)) {
        cout << "Read bitmap image: Image size mismatch (expecting 3 bytes per pixel) " << endl;
        return false;
    }
    // don't bother checking biPlanes
    if (bmih.biBitCount !=  24) {
        cout << "Read bitmap image: Expecting 3 bytes per pixel for greyscale";
        cout << endl;
        return false;
    }
    // don't bother checking biConmpression, biSizeImage, biXPelsPerMeter,
    // biYPelsPerMeter, biClrUsed, or biClrImportant

    // allocate space;
    unsigned char byte;
    data =  new unsigned char[w*h];
    if (!data)  {
        cout << "Read bitmap file: error allocating image data array" << endl;
        return false;
    }
    
    int i, j;
    // Read image data.  Each byte in each row must be written 3 times
    for (i = h-1; i >= 0; i--)  {
        for (j =  0; j < w; j++) {
            if (!read_three_copies(bmpfile, byte)) {
                 cout << "Read bitmap file: Error reading image data" << endl;
                 return false;
            }
            data[i*w+j] = byte;
        }
    }
    bmpfile.close();
    return data;
}


